#
# ht.pl -- HTTP abstractions
#
# Deal with CGI and do a few extremely common operations.  Nothing
# major, just niggling little bits.
#
# XXX - needs to handle the HEAD method! (only HTTP headers and
#	no document body, whatever that means).  Should <HEAD></HEAD>
#	stuff be included?  Probably would be nice if Title: and other
#	stuff isn't in the HTTP headers.
#
# By default, we assume that the header is not parsed.
# Do "$ht'ph = 1" to turn that off.
#
# Has the side effect of setting $0 to the request being processed.
# This may not be nice if you have passwords or something in there,
# but it sure helps in seeing what's going on.  Sadly, it doesn't
# seem to work with my version of perl on the HPs.  I've made a
# path that fixes that, though.
#
# I'm thinking of reworking this so that there are calls to set
# the values of headers in the output instead of specifying them
# all in a single function call.  Either way, this library could
# change drastically in the future.
#
# Author:
#   George Phillips <phillips@cs.ubc.ca>
#   Department of Computer Science
#   University of British Columbia

$0 = "$ENV{'REQUEST_METHOD'} $ENV{'SCRIPT_NAME'}$ENV{'PATH_INFO'}";

package ht;

$ph = 0;

# Protocol boilerplate

# Generic document header routine.  Called directly for binary data
# such as GIF or JPEG; but use support routines for other documents

sub start {
	local($content, $length) = @_;

	&mimehead('200', '', $content, $length);

	$doc_content = $content;
}

sub end {
	print "</BODY>\n"
		if $doc_content eq 'html' ||
		   $doc_content eq 'text/html' ||
		   $doc_content eq '';
}

sub html {
	&start('html');
	print "<HEAD>\n";
	print "<TITLE>$_[0]</TITLE>\n";
	print "</HEAD>\n";
	print "<BODY>\n";
}

sub ht_html { 
	&html($_[0]);
	print "<H1>$_[0]</H1>\n";
}

sub redirect {
	local($to) = shift;

	&mimehead('301', 'Redirect', 'html', -1, "Location: $to\r\n");
	print "<HEAD><TITLE>Document has moved</TITLE></HEAD>\n";
	print "<H1>Document has moved</H1>\n";
	print "Try opening <A HREF=\"$to\">", &esc($to), "</A>.\n";
	exit 0;
}

sub error {
	local($code, $brief, $long) = @_;

	$brief = &mimehead($code, $brief, 'html');
	print "<HEAD><TITLE>Error: $brief</TITLE></HEAD>\n";
	print "<H1>Error: $brief</H1>\n";
	print "$long\n";
	exit 1;
}

sub mimehead {
	local($code, $userinfo, $content, $length, $extra) = @_;
	local($mimecontent);

	$code = '200' if $code eq 'ok';
	$code = '400' if $code eq 'badreq';
	$code = '404' if $code eq 'notfound';
	$code = '500' if $code eq 'internal';

	if ($userinfo eq '') {
		$userinfo = 'Document follows' if $code eq '200';
		$userinfo = 'Internal error' if $code eq '500';
		$userinfo = 'Document not found' if $code eq '404';
	}

	$code =~ /^\d+$/ ||
		&error('500', '',
			"This server really blew it by giving '$code' as an response code.
			<P>Oh well.  The original brief message was:<P>
			<H2>$brief</H2>\n");

	$mimecontent = 'application/octet-stream';
	$mimecontent = 'text/plain' if $content eq 'text';
	$mimecontent = 'text/html' if $content eq 'html' || $content eq '';
	$mimecontent = 'image/gif' if $content eq 'gif';
	$mimecontent = 'image/jpeg' if $content eq 'jpeg';
	$mimecontent = $content if $content =~ m'/';

	if ($ENV{'SERVER_PROTOCOL'} ne 'HTTP/0.9') {
		print "HTTP/1.0 $code $userinfo\r\n" if !$ph;
		print "MIME-Version: 1.0\r\n" if !$ph;
		print $extra;
		print "Content-Type: $mimecontent\r\n";
		print "Content-Length: $length\r\n" if $length > 0;
		$| = 1 if &contentbinary($content);
		print "\r\n";
		$| = 0 if &contentbinary($content);
	}

	# XXX urg; a hack for &ht'error
	$userinfo;
}

sub contentbinary {
	$_[0] ne 'text/html' && $_[0] ne 'text/plain';
}

# escape &, < and > for inclusion in HTML

sub esc {
	$_[0] =~ s/\&/\&amp;/g;
	$_[0] =~ s/\</\&lt;/g;
	$_[0] =~ s/\>/\&gt;/g;

	$_[0];
}

sub unesc_query {
	$_[0] =~ tr/+/ /;
	$_[0] =~ s/\%(..)/pack('H2', $1)/eg;
	$_[0];
}

# XXX we'll probably be wanting a hex escaper (and unescaper) soon enough

1;
